﻿#region BSD License
/* 
Copyright (c) 2010, NETFx
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

* Neither the name of Clarius Consulting nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#endregion
using System;

/// <summary>
/// Provides conversion of DateTime and DateTimeOffset into an epoch-relative number value (total seconds).
/// See Unix Epoch in Wikipedia for more information on why this might be needed. 
/// Typical uses include using this simplified representation as an expiration time for a token, password or verification code.
/// </summary>
/// <nuget id="netfx-System.DateTimeEpochExtensions"/>
internal static class DateTimeEpochExtensions
{
	/// <summary>
	/// Converts the given date value to epoch time.
	/// </summary>
	/// <param name="dateTime" this="true">The data to convert</param>
	public static long ToEpochTime(this DateTime dateTime)
	{
		var date = dateTime.ToUniversalTime();
		var ts = date - new DateTime(1970, 1, 1, 0, 0, 0, 0, DateTimeKind.Utc);

		return Convert.ToInt64(ts.TotalSeconds);
	}

	/// <summary>
	/// Converts the given date value to epoch time.
	/// </summary>
	/// <param name="dateTime" this="true">The data to convert</param>
	public static long ToEpochTime(this DateTimeOffset dateTime)
	{
		var date = dateTime.ToUniversalTime();
		var ts = date - new DateTimeOffset(1970, 1, 1, 0, 0, 0, TimeSpan.Zero);

		return Convert.ToInt64(ts.TotalSeconds);
	}

	/// <summary>
	/// Converts the given epoch time to a <see cref="DateTime"/> with <see cref="DateTimeKind.Utc"/> kind.
	/// </summary>
	/// <param name="secondsSince1970" this="true">The seconds to convert</param>
	public static DateTime ToDateTimeFromEpoch(this long secondsSince1970)
	{
		return new DateTime(1970, 1, 1, 0, 0, 0, 0, DateTimeKind.Utc).AddSeconds(secondsSince1970);
	}

	/// <summary>
	/// Converts the given epoch time to a UTC <see cref="DateTimeOffset"/>.
	/// </summary>
	/// <param name="secondsSince1970" this="true">The seconds to convert</param>
	public static DateTimeOffset ToDateTimeOffsetFromEpoch(this long secondsSince1970)
	{
		return new DateTimeOffset(1970, 1, 1, 0, 0, 0, TimeSpan.Zero).AddSeconds(secondsSince1970);
	}
}