﻿#region BSD License
/* 
Copyright (c) 2011, NETFx
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this list 
  of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice, this 
  list of conditions and the following disclaimer in the documentation and/or other 
  materials provided with the distribution.

* Neither the name of Clarius Consulting nor the names of its contributors may be 
  used to endorse or promote products derived from this software without specific 
  prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED 
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR 
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH 
DAMAGE.
*/
#endregion
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.IO.Compression;

/// <summary>
/// An implementation of the core NETFx ISerializer interface that provides stream compression/decompression on top of another serializer.
/// </summary>
///	<nuget id="netfx-System.GzipSerializer" />
partial class GzipSerializer : ISerializer
{
	private ISerializer innerSerializer;

	/// <summary>
	/// Initializes a new instance of the <see cref="GzipSerializer"/> class 
	/// with the underlying serializer to use over the compressed stream.
	/// </summary>
	/// <param name="innerSerializer">The inner serializer.</param>
	public GzipSerializer(ISerializer innerSerializer)
	{
		Guard.NotNull(() => innerSerializer, innerSerializer);

		this.innerSerializer = innerSerializer;
	}

	/// <summary>
	/// Deserializes an object graph of type <typeparamref name="T"/> 
	/// from the given compressed stream.
	/// </summary>
	/// <typeparam name="T">The type of object graph to deserialize.</typeparam>
	/// <param name="stream">The serialized binary stream.</param>
	public virtual T Deserialize<T>(Stream stream)
	{
		Guard.NotNull(() => stream, stream);

		using (var decompress = new DeflateStream(stream, CompressionMode.Decompress, true))
		{
			return this.innerSerializer.Deserialize<T>(decompress);
		}
	}

	/// <summary>
	/// Serializes the given object graph as a compressed binary stream.
	/// </summary>
	/// <typeparam name="T">The type of object graph to serialize, inferred by the 
	///  compiler from the passed-in <paramref name="graph"/>.</typeparam>
	/// <param name="stream">The stream to serialize into.</param>
	/// <param name="graph">The object graph to serialize.</param>
	public virtual void Serialize<T>(Stream stream, T graph)
	{
		Guard.NotNull(() => stream, stream);
		Guard.NotNull(() => graph, graph);

		using (var compress = new DeflateStream(stream, CompressionMode.Compress, true))
		{
			this.innerSerializer.Serialize<T>(compress, graph);
		}
	}
}